/*
  Copyright 2019 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.auth

import android.app.Activity
import android.content.Context
import android.content.Intent
import android.net.Uri
import android.os.Bundle
import android.os.Handler
import android.os.Looper
import android.os.ResultReceiver
import android.util.Log
import com.kakao.sdk.common.util.KakaoJson
import com.kakao.sdk.common.KakaoSdk
import com.kakao.sdk.common.model.*
import com.kakao.sdk.common.util.IntentResolveClient
import java.net.HttpURLConnection

/**
 * OAuth 2.0 Spec 의 authorization code 를 발급받기 위해 사용되는 클라이언트.
 */
class AuthCodeClient (
    private val authApiClient: AuthApiClient = AuthApiClient.instance,
    private val intentResolveClient: IntentResolveClient = IntentResolveClient.instance
) {

    private val CAPRI_LOGGED_IN_ACTIVITY = "com.kakao.talk.intent.action.CAPRI_LOGGED_IN_ACTIVITY"

    /**
     * 카카오톡 간편로그인 가능 여부(설치 여부)를 검사합니다.
     */
    fun isTalkLoginAvailable(context: Context): Boolean =
        intentResolveClient.resolveTalkIntent(context, baseTalkLoginIntent()) != null

    /**
     * 카카오톡 간편 로그인을 통하여 authorization code 를 발급 받는다.
     *
     * @param context instance for starting an activity
     */
    @JvmOverloads
    fun authorizeWithKakaoTalk(
        context: Context,
        requestCode: Int = DEFAULT_REQUEST_CODE,
        channelPublicIds: List<String>? = null,
        serviceTerms: List<String>? = null,
        autoLogin: Boolean? = null,
        redirectUri: String = "kakao${KakaoSdk.applicationContextInfo.appKey}://oauth",
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        kaHeader: String = KakaoSdk.applicationContextInfo.kaHeader,
        callback: (code: String?, error: Throwable?) -> Unit
    ) {
        val extras = Bundle().apply {
            channelPublicIds?.let { putString(Constants.CHANNEL_PUBLIC_ID, channelPublicIds.joinToString(",")) }
            serviceTerms?.let { putString(Constants.SERVICE_TERMS, serviceTerms.joinToString(",")) }
            autoLogin?.let { putString(Constants.AUTO_LOGIN, it.toString()) }
        }
        val talkIntent = talkLoginIntent(clientId, redirectUri, kaHeader, extras)
        val resolvedIntent = intentResolveClient.resolveTalkIntent(context, talkIntent)
        if (resolvedIntent == null) {
            val error = ClientError(ClientErrorCause.NotSupported, "KakaoTalk not installed")
            callback(null, error)
            return
        }
        context.startActivity(
            Intent(context, TalkAuthCodeActivity::class.java)
                .putExtra(Constants.KEY_LOGIN_INTENT, resolvedIntent)
                .putExtra(Constants.KEY_REQUEST_CODE, requestCode)
                .putExtra(Constants.KEY_RESULT_RECEIVER, resultReceiver(callback))
        )
    }

    /**
     * Custom Tabs (기본 브라우저) 를 사용하여 authorization code 를 발급 받는다.
     *
     * @param context instance for starting an activity
     */
    @JvmOverloads
    fun authorizeWithKakaoAccount(
        context: Context,
        channelPublicIds: List<String>? = null,
        serviceTerms: List<String>? = null,
        autoLogin: Boolean? = null,
        redirectUri: String = "kakao${KakaoSdk.applicationContextInfo.appKey}://oauth",
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        kaHeader: String = KakaoSdk.applicationContextInfo.kaHeader,
        callback: (code: String?, error: Throwable?) -> Unit
    ) {
        val fullUri = UriUtility.authorizeUri(clientId, null, redirectUri, null, kaHeader, channelPublicIds, serviceTerms, autoLogin)
        context.startActivity(authCodeIntent(context, fullUri, redirectUri, resultReceiver(callback)))
    }

    /**
     * 유저의 동의항목을 업데이트할 수 있는 authorization code 를 발급받는다.
     *
     * @param context [Context] instance for starting an activity
     * @param scopes list of scopes to be updated
     */
    @JvmOverloads
    fun authorizeWithNewScopes(
        context: Context,
        scopes: List<String>,
        channelPublicIds: List<String>? = null,
        serviceTerms: List<String>? = null,
        autoLogin: Boolean? = null,
        redirectUri: String = "kakao${KakaoSdk.applicationContextInfo.appKey}://oauth",
        clientId: String = KakaoSdk.applicationContextInfo.appKey,
        kaHeader: String = KakaoSdk.applicationContextInfo.kaHeader,
        callback: (code: String?, error: Throwable?) -> Unit
    ) {
        authApiClient.agt(clientId) { response, error ->
            response?.let {
                val fullUri = UriUtility.authorizeUri(clientId, response.agt, redirectUri, scopes, kaHeader, channelPublicIds, serviceTerms, autoLogin)
                context.startActivity(authCodeIntent(context, fullUri, redirectUri, resultReceiver(callback)))
                return@agt
            }
            callback(null, error)
        }
    }

    @JvmSynthetic
    internal fun resultReceiver(
        callback: (code: String?, error: Throwable?) -> Unit
    ): ResultReceiver =
        object: ResultReceiver(Handler(Looper.getMainLooper())) {
            override fun onReceiveResult(resultCode: Int, resultData: Bundle?) {
                Log.d("***** AUTH CODE RESULT", "$resultData")
                if (resultCode == Activity.RESULT_OK) {
                    val uri = resultData?.getParcelable<Uri>(Constants.KEY_URL)
                    val code = uri?.getQueryParameter(Constants.CODE)
                    if (code != null) {
                        callback(code, null)
                    } else {
                        val error = uri?.getQueryParameter(Constants.ERROR) as String
                        val errorDescription = uri.getQueryParameter(Constants.ERROR_DESCRIPTION) as String
                        callback(null, OAuthError(
                            HttpURLConnection.HTTP_MOVED_TEMP,
                            KakaoJson.fromJson(
                                error,
                                AuthErrorCause::class.java
                            )
                                ?: AuthErrorCause.Unknown,
                            AuthErrorResponse(error, errorDescription)
                        ))
                    }

                } else if (resultCode == Activity.RESULT_CANCELED) {
                    var error = resultData?.getSerializable(Constants.KEY_EXCEPTION) as KakaoSdkError
                    callback(null, error)
                } else {
                    var error = IllegalArgumentException("Unknown resultCode in RxAuthCodeClient#onReceivedResult()")
                    callback(null, error)
                }
            }
        }

    @JvmSynthetic
    internal fun baseTalkLoginIntent(): Intent =
        Intent().setAction(CAPRI_LOGGED_IN_ACTIVITY).addCategory(Intent.CATEGORY_DEFAULT)

    @JvmSynthetic
    internal fun talkLoginIntent(
        clientId: String,
        redirectUri: String,
        kaHeader: String,
        extras: Bundle
    ): Intent =
        baseTalkLoginIntent()
            .putExtra(Constants.EXTRA_APPLICATION_KEY, clientId)
            .putExtra(Constants.EXTRA_REDIRECT_URI, redirectUri)
            .putExtra(Constants.EXTRA_KA_HEADER, kaHeader)
            .putExtra(Constants.EXTRA_EXTRAPARAMS, extras)

    @JvmSynthetic
    internal fun authCodeIntent(
        context: Context,
        fullUri: Uri,
        redirectUri: String,
        resultReceiver: ResultReceiver
    ): Intent =
        Intent(context, AuthCodeCustomTabsActivity::class.java)
            .putExtra(Constants.KEY_BUNDLE, Bundle().apply {
                putParcelable(Constants.KEY_RESULT_RECEIVER, resultReceiver)
                putParcelable(Constants.KEY_FULL_URI, fullUri)
                putString(Constants.KEY_REDIRECT_URI, redirectUri)
            })
            .addFlags(Intent.FLAG_ACTIVITY_NEW_TASK)    // (동적동의 시 SDK가 갖고 있던) Application Context 로 Activity를 띄우기 위해 설정

    companion object {

        /**
         * 간편한 API 호출을 위해 기본 제공되는 singleton 객체
         */
        @JvmStatic
        val instance by lazy { AuthCodeClient() }

        /**
         * 카카오톡 간편로그인 startActivity 에서 사용되는 requestCode 기본 값
         */
        const val DEFAULT_REQUEST_CODE: Int = 10012
    }
}